module     uussbb_dipoles
   use precision, only: ki
   use uussbb_color, only: gammaF, gammaA, &
       & CF, CA, numcs, KF, KA
   use uussbb_model
   use uussbb_kinematics, only: num_legs, dotproduct
   implicit none

   real(ki), parameter :: pi = &
   & 3.1415926535897932384626433832795028841971693993751_ki
   real(ki), parameter :: pi3  = pi**2 / 3.0_ki
   real(ki), parameter :: pi6  = pi**2 / 6.0_ki
   real(ki), parameter :: pi12 = pi**2 / 12.0_ki

   real(ki), dimension(0), parameter :: mF_null = 0

   real(ki) :: Qaux_2 = 1.0_ki
   real(ki) :: kappa  = 2.0_ki/3.0_ki

   private :: V_SING, V_NSING_A, V_NSING_F, GAMMA_F, GAMMA_A
contains


   function V_SING(s_jk, Q_jk, m_j, m_k, v_jk, rho, rho_j_2, rho_k_2)
      implicit none
      real(ki), intent(in) :: s_jk, Q_jk, m_j, m_k
      real(ki), intent(in) :: v_jk, rho, rho_j_2, rho_k_2
      type(form_factor) :: V_SING
      real(ki) :: tmp, tmp2, tmp3

      if (m_j .gt. 0.0_ki) then
              if(m_k .gt. 0.0_ki) then
                      ! (6.20), case 1, both massive
                      tmp  = log(rho)
                      tmp2 = log(rho_j_2)
                      tmp3 = log(rho_k_2)
                      V_SING = form_factor(0.0_ki, tmp, &
                      & - 0.25_ki * (tmp2*tmp2+tmp3*tmp3) - pi6 &
                      & + tmp*log(Q_jk/s_jk)) / v_jk
              else
                      ! (6.20), case 2, only m_j massive
                      tmp  = log(m_j*m_j/s_jk)
                      tmp3 = log(m_j*m_j/Q_jk)
                      V_SING = form_factor(0.5_ki, 0.5_ki * tmp, &
                        & - 0.25_ki * tmp * tmp - pi12 &
                        & - 0.5_ki * (tmp + tmp3) * log(s_jk/Q_jk) )
              end if
      else
              if(m_k .gt. 0.0_ki) then
                      ! (6.20), case 2, only m_k massive
                      tmp  = log(m_k*m_k/s_jk)
                      tmp3 = log(m_k*m_k/Q_jk)
                      V_SING = form_factor(0.5_ki, 0.5_ki * tmp, &
                        & - 0.25_ki * tmp * tmp - pi12 &
                        & - 0.5_ki * (tmp + tmp3) * log(s_jk/Q_jk) )
              else
                      ! (6.20), case 3, both massless
                      V_SING = form_factor(1.0_ki, 0.0_ki, 0.0_ki)
              end if
      end if
   end  function V_SING

   function V_NSING_F(s_jk, Q_jk, m_j, m_k, v_jk, &
         rho, rho_j_2, rho_k_2, mF)
      implicit none
      real(ki), intent(in) :: s_jk, Q_jk, m_j, m_k
      real(ki), intent(in) :: v_jk, rho, rho_j_2, rho_k_2
      real(ki), dimension(:), intent(in) :: mF
      complex(ki) :: V_NSING_F
      real(ki) :: tmp, tmp2, tmp3, tmp4

      if (m_j .gt. 0.0_ki) then
              if(m_k .gt. 0.0_ki) then
                      ! (6.21), both massive
                      tmp  = rho*rho
                      tmp2 = sqrt(Q_jk)
                      tmp3 = (tmp2 - m_k) / tmp2
                      tmp4 = m_j*m_j

                      V_NSING_F = gammaF/CF * log(s_jk/Q_jk) &
                      & + (log(tmp)*log(1.0_ki+tmp) &
                      &  + 2.0_ki * zdilog(tmp, +1.0_ki) &
                      &  - zdilog(1.0_ki - rho_j_2, +1.0_ki) &
                      &  - zdilog(1.0_ki - rho_k_2, +1.0_ki) - pi6) / v_jk&
                      & + log(tmp3) - 2.0_ki * log(tmp3*tmp3-tmp4/Q_jk) &
                      & - 2.0_ki * tmp4/s_jk * log(m_j/(tmp2 - m_k)) &
                      & - m_k/(tmp2 - m_k) &
                      & + (m_k + m_k) * (m_k + m_k - tmp2) / s_jk &
                      & + pi*pi/2.0_ki
              else
                      ! (6.22)
                      tmp  = m_j*m_j
                      tmp2 = s_jk/Q_jk
                      V_NSING_F = (gammaF/CF - 2.0_ki) * log(tmp2) &
                      & + pi6 - zdilog(tmp2, +1.0_ki) &
                      & - tmp/s_jk * log(tmp/Q_jk)
              end if
      else
              if(m_k .gt. 0.0_ki) then
                      ! (6.23)
                      tmp = sqrt(Q_jk)
                      V_NSING_F = gammaF/CF * (log(s_jk/Q_jk) &
                      & - 2.0_ki * (log((tmp - m_k) / tmp) &
                      & + m_k/(tmp + m_k))) + pi6 - zdilog(s_jk/Q_jk, +1.0_ki)
              else
                      ! V_q^(NS)(s, 0, 0) = 0
                      V_NSING_F = 0.0_ki
              end if
      end if
      ! print*, "V_NSING_F=", V_NSING_F
   end  function V_NSING_F

   function V_NSING_A(s_jk, q_jk, m_k, v_jk, mF, kappa)
      implicit none
      real(ki), intent(in) :: s_jk, Q_jk, m_k, v_jk, kappa
      real(ki), dimension(:), intent(in) :: mF
      complex(ki) :: V_NSING_A
      real(ki) :: tmp, tmp2, tmp3, tmp4, rho1, rho2
      real(ki) :: m, rkappa, rh_2, rh_3
      integer :: i


      rkappa = (kappa - 2.0_ki/3.0_ki) * m_k*m_k/s_jk
      if(m_k .gt. 0.0_ki) then
              ! (6.24)
              tmp  = sqrt(Q_jk)
              tmp2 = 0.0_ki
              tmp3 = 0.0_ki
              do i = lbound(mF,1), ubound(mF,1)
                 m = mF(i)
                 tmp4 = m*m
                 tmp2 = tmp2 + log(tmp4/Qaux_2)
                 if (s_jk > 4.0_ki * m * (m + m_k)) then
                         rho1 = (tmp - m)
                         rho1 = sqrt(1.0_ki-4.0_ki*tmp4/(rho1*rho1))
                         rh_2 = rho1 * rho1
                         rh_3 = rho1 * rh_2
                         rho2 = sqrt(1.0_ki-4.0_ki*tmp4/(Q_jk - m_k*m_k))
                         tmp3 = tmp3 + log((tmp - m_k)/tmp) &
                         & + m_k * rh_3/(tmp+m_k) &
                         & + log((1.0_ki + rho1)/2.0_ki) &
                         & - rho1/3.0_ki*(3.0_ki+rh_2) &
                         & - 0.5_ki * log(tmp4/Q_jk)
                         if(rkappa .ne. 0.0_ki) then
                           tmp3 = tmp3 + rkappa * 2.0_ki * TR / CA * ( &
                           & rho2*rho2*rho2 * log((rho2-rho1)/(rho2+rho1)) &
                           & - log((1.0_ki - rho1)/(1.0_ki + rho1)) &
                           & - 8.0_ki * rho1 * tmp4 / s_jk)
                         end if
                 end if
              end do

              V_NSING_A = gammaA/CA * (&
              &   log(s_jk/Q_jk) &
              & - 2.0_ki * log((tmp - m_k)/tmp) &
              & - (m_k + m_k) / (tmp + m_k) &
              & ) &
              & + pi6 - zdilog(s_jk/Q_jk, +1.0_ki) &
              & + 2.0_ki/3.0_ki * TR/CA * (tmp2 + tmp3 + tmp3)
              if(rkappa .ne. 0.0_ki) then
                 V_NSING_A = V_NSING_A &
                 & + rkappa * (2.0_ki * TR * NF / CA - 1.0_ki) * &
                 &     log((m_k + m_k) / (tmp + m_k))
              end if

      else
              ! (6.26)
              tmp  = sqrt(Q_jk)
              tmp2 = 0.0_ki
              tmp3 = 0.0_ki
              do i = lbound(mF,1), ubound(mF,1)
                 m = mF(i)
                 tmp4 = m*m
                 tmp2 = tmp2 + log(tmp4/Qaux_2)
                 if (s_jk > 4.0_ki * m * (m + m_k)) then
                         rho1 = (tmp - m)
                         rho1 = sqrt(1.0_ki-4.0_ki*tmp4/(rho1*rho1))
                         tmp3 = tmp3 + log((1.0_ki + rho1)/2.0_ki) &
                         & - rho1/3.0_ki*(3.0_ki + rho1*rho1) &
                         & - 0.5_ki * log(tmp4/s_jk)
                 end if
              end do

              V_NSING_A = 2.0_ki/3.0_ki * TR/CA * (tmp2 + tmp3 + tmp3)
      end if
      ! print*, "V_NSING_A=", V_NSING_A
   end  function V_NSING_A

   function GAMMA_A(mF)
      ! Equation (6.27)
      implicit none
      real(ki), dimension(:), intent(in) :: mF
      type(form_factor) :: GAMMA_A
      real(ki) :: m, tmp
      integer :: i

      tmp = 0.0_ki
      do i = lbound(mF,1), ubound(mF,1)
         m = mF(i)
         tmp = tmp + log(m*m/Qaux_2)
      end do
      tmp = - 2.0_ki/3.0_ki * TR * tmp
      GAMMA_A = form_factor(0.0_ki, gammaA, tmp)
   end  function GAMMA_A

   function GAMMA_F(mu_sq, mq)
      implicit none
      real(ki), intent(in) :: mu_sq, mq
      type(form_factor) :: GAMMA_F

      if (mq .gt. 0.0_ki) then
              ! (6.29)
              GAMMA_F = CF * form_factor(0.0_ki, 1.0_ki, &
              & 0.5_ki * log(mq*mq/mu_sq) - 2.0_ki)
      else
              ! (6.28)
              GAMMA_F = form_factor(0.0_ki, gammaF, 0.0_ki)
      end if
   end  function GAMMA_F

   function lambda(x, y, z)
      implicit none
      real(ki), intent(in) :: x, y, z
      real(ki) :: lambda, tmp

      lambda = x - y
      tmp    = x + y
      lambda = lambda * lambda
      lambda = lambda + z*(z - tmp - tmp)
   end  function lambda
end module uussbb_dipoles
